theory executions imports Main begin

typedecl event
  
definition "imm r \<equiv> r - (r O r\<^sup>+)"
  
definition lift ("\<lfloor> _ \<rfloor>") where "lift s \<equiv> {(x,x)| x. x \<in> s}"


(* lifting from events to transactions *)
fun weaklift :: "event rel \<times> event rel \<Rightarrow> event rel"
  where "weaklift(r, t) = t O (r - t) O t"
fun stronglift :: "event rel \<times> event rel \<Rightarrow> event rel"
  where "stronglift(r, t) = t^= O (r - t) O t^="
  
record execution = 
  E :: "event set"      -- "all events"
  R :: "event set"      -- "read events"
  W :: "event set"      -- "write events"
  po :: "event rel"     -- "program order"
  ad :: "event rel"     -- "address dependency"
  cd :: "event rel"     -- "control dependency"
  dd :: "event rel"     -- "data dependency"
  rmw :: "event rel"    -- "components of an RMW operation"
  sloc :: "event rel"   -- "same-location equivalence"
  sthd :: "event rel"   -- "same-thread equivalence"
  stxn :: "event rel"   -- "same-transaction equivalence"
  F :: "event set"      -- "(C++) fence events"
  A :: "event set"      -- "(C++) atomic events"
  ACQ :: "event set"    -- "(C++) acquiring events"
  REL :: "event set"    -- "(C++) releasing events"
  SC :: "event set"     -- "(C++) SC events"
  mfence :: "event rel" -- "(x86) MFENCE"
  isync :: "event rel"  -- "(Power) ISYNC"
  lwsync :: "event rel" -- "(Power) LWSYNC"
  sync :: "event rel"   -- "(Power) SYNC"
  SCACQ :: "event set"  -- "(ARM) acquiring events"
  SCREL :: "event set"  -- "(ARM) releasing events"
  isb :: "event rel"    -- "(ARM) ISB"
  dmbld :: "event rel"  -- "(ARM) DMB.LD"
  dmbst :: "event rel"  -- "(ARM) DMB.ST"
  dmb :: "event rel"    -- "(ARM) DMB"
  
inductive wf_exec where
  "\<lbrakk> 
  (* Event sets are well-formed *)
  R X \<subseteq> E X;
  W X \<subseteq> E X;
  F X \<subseteq> E X;
  A X \<subseteq> E X;
  ACQ X \<subseteq> E X;
  REL X \<subseteq> E X;
  SC X \<subseteq> E X;
  SCACQ X \<subseteq> E X;
  SCREL X \<subseteq> E X;
  (* Event relations are well-formed *)
  po X \<subseteq> E X \<times> E X;
  ad X \<subseteq> E X \<times> E X;
  cd X \<subseteq> E X \<times> E X;
  dd X \<subseteq> E X \<times> E X;
  sloc X \<subseteq> E X \<times> E X;
  sthd X \<subseteq> E X \<times> E X;
  stxn X \<subseteq> E X \<times> E X;
  rmw X \<subseteq> E X \<times> E X;
  rf \<subseteq> E X \<times> E X;
  co \<subseteq> E X \<times> E X;
  mfence X \<subseteq> E X \<times> E X;
  isync X \<subseteq> E X \<times> E X;
  lwsync X \<subseteq> E X \<times> E X;
  sync X \<subseteq> E X \<times> E X;
  isb X \<subseteq> E X \<times> E X;
  dmbld X \<subseteq> E X \<times> E X;
  dmbst X \<subseteq> E X \<times> E X;
  dmb X \<subseteq> E X \<times> E X;
  (* Reads, writes, and fences are disjoint *)
  R X \<inter> W X = {};
  (R X \<union> W X) \<inter> F X = {};
  (* Program-order is intra-thread *)
  po X \<subseteq> sthd X;
  (* Program-order is a strict partial order *)
  acyclic (po X); trans (po X);
  (* Address dependencies are read to read/write *)
  ad X \<subseteq> (R X \<times> (R X \<union> W X)) \<inter> po X;
  (* Control dependencies are read to anything *)
  cd X \<subseteq> (R X \<times> E X) \<inter> po X;
  (* Data dependencies are read to write *)
  dd X \<subseteq> (R X \<times> W X) \<inter> po X;
  (* RMWs relate reads to writes in program order *)
  rmw X \<subseteq> (R X \<times> W X) \<inter> sloc X \<inter> po X;
  (* Same-thread is an equivalence over all events *)
  equiv (E X) (sthd X);
  (* Same-location is an equivalence over reads/writes *)
  equiv (R X \<union> W X) (sloc X);
  (* Reads-from is on the same location *)
  rf \<subseteq> sloc X;
  (* Reads-from is write-to-read *)
  rf \<subseteq> W X \<times> R X;
  (* Reads-from is left-unique *)
  \<forall>w1 w2 r. (w1,r) \<in> rf \<longrightarrow> (w2,r) \<in> rf \<longrightarrow> w1 = w2;
  (* Coherence-order is a strict partial order *)
  acyclic co; trans co;
  (* Coherence-order is a union, over all locations x, of
     strict total orders on writes to x *)
  co \<union> co^-1 = (W X \<times> W X) \<inter> (sloc X) - Id;
  (* Transactions are intra-thread *)
  stxn X \<subseteq> sthd X;
  (* Same-transaction is a partial equivalence on events *)
  sym (stxn X); trans (stxn X);
  (* Transactions are contiguous *)
  (((po X O po X) \<inter> stxn X) O (po X)\<inverse>) \<inter> po X \<subseteq> stxn X
  \<rbrakk> \<Longrightarrow> 
  wf_exec(X,rf,co)"
  
  
inductive wf_exec_cpp where
  "\<lbrakk> 
  wf_exec(X,rf,co);
  mfence X = {};
  isync X = {}; lwsync X = {}; sync X = {};
  SCACQ X = {}; SCREL X = {}; isb X = {};
  dmbld X = {}; dmbst X = {}; dmb X = {};
  (* (C++) Acquires, releases, and SC events are all atomic *)
  ACQ X \<union> REL X \<union> SC X \<subseteq> A X;
  (* (C++) Fences are atomic *)
  F X \<subseteq> A X;
  (* (C++) SC reads can acquire *)
  R X \<inter> SC X \<subseteq> ACQ X;
  (* (C++) Only reads and fences can acquire *)
  ACQ X \<subseteq> R X \<union> F X;
  (* (C++) SC writes can release *)
  W X \<inter> SC X \<subseteq> REL X;
  (* (C++) Only writes and fences can release *)
  REL X \<subseteq> W X \<union> F X;
  (* (C++) SC fences can acquire and release *)
  F X \<inter> SC X \<subseteq> ACQ X \<inter> REL X;
  (* RMWs are consecutive and do not straddle txn boundaries *)
  rmw X \<subseteq> imm(po X);
  rmw X \<subseteq> stxn X \<union> ((E X - Domain(stxn X)) \<times> (E X - Domain(stxn X)))
  \<rbrakk> \<Longrightarrow> 
  wf_exec_cpp(X,rf,co)"
  
inductive wf_exec_x86 where
  "\<lbrakk> 
  wf_exec(X,rf,co);
  ACQ X = {}; REL X = {}; SC X = {}; F X = {};
  isync X = {}; lwsync X = {}; sync X = {};
  SCACQ X = {}; SCREL X = {}; isb X = {};
  dmbld X = {}; dmbst X = {}; dmb X = {};
  (* Control dependencies in assembly *)
  cd X O po X \<subseteq> cd X;
  (* RMWs are consecutive and do not straddle txn boundaries *)
  rmw X \<subseteq> imm(po X);
  rmw X \<subseteq> stxn X \<union> ((E X - Domain(stxn X)) \<times> (E X - Domain(stxn X)))
  \<rbrakk> \<Longrightarrow> 
  wf_exec_x86(X,rf,co)"
  
inductive wf_exec_power where
  "\<lbrakk> 
  wf_exec(X,rf,co);
  ACQ X = {}; REL X = {}; SC X = {}; F X = {};
  mfence X = {};
  SCACQ X = {}; SCREL X = {}; isb X = {};
  dmbld X = {}; dmbst X = {}; dmb X = {};
  (* Control dependencies in assembly *)
  cd X O po X \<subseteq> cd X;
  (* Full fence implies lightweight fence *)
  sync X \<subseteq> lwsync X
  \<rbrakk> \<Longrightarrow> 
  wf_exec_power(X,rf,co)"
  
inductive wf_exec_arm where
  "\<lbrakk> 
  wf_exec(X,rf,co);
  ACQ X = {}; REL X = {}; SC X = {}; F X = {};
  mfence X = {};
  isync X = {}; lwsync X = {}; sync X = {};
  (* Control dependencies in assembly *)
  cd X O po X \<subseteq> cd X;
  (* Strong fences imply weaker fences *)
  dmb X \<subseteq> dmbst X;
  dmb X \<subseteq> dmbld X;
  (* Acquires are reads, releases are writes *)
  SCACQ X \<subseteq> R X;
  SCREL X \<subseteq> W X
  \<rbrakk> \<Longrightarrow> 
  wf_exec_arm(X,rf,co)"
  
definition "into X r \<equiv> ((E X \<times> E X) - r) O r"
definition "outof X r \<equiv> r O ((E X \<times> E X) - r)"
  
type_synonym witnessed_execution = 
  "execution \<times> event rel \<times> event rel"
  
record memory_model =
  consistent :: "witnessed_execution set"
  racefree :: "witnessed_execution set"
   
definition "allowed M P \<equiv> 
  let cands = {(X,rf,co) :: witnessed_execution | X rf co. X \<in> P} in
  if consistent M \<inter> cands \<subseteq> racefree M 
  then fst ` (consistent M \<inter> cands) else UNIV"
  

  
(* from-read *)
fun fr :: "witnessed_execution \<Rightarrow> event rel"
  where "fr (X,rf,co) = (R X \<times> W X) \<inter> sloc X - (rf\<inverse> O (co\<inverse>)^*)"
    
(* communication *)
fun com :: "witnessed_execution \<Rightarrow> event rel" 
  where "com (X,rf,co) = rf \<union> co \<union> fr(X,rf,co)"
    
(* inter-thread from-read *)
fun fre :: "witnessed_execution \<Rightarrow> event rel"
  where "fre (X,rf,co) = fr (X,rf,co) - sthd X"
    
(* inter-thread reads-from *)
fun rfe :: "witnessed_execution \<Rightarrow> event rel"
  where "rfe (X,rf,co) = rf - sthd X"
  
end
  